// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2012 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
#if NET_3_5
using System.Linq;
#else
using Smdn.Collections;
#endif
using System.Text;

using Smdn.IO;

namespace Smdn.Formats.Mime {
  public class MimeMessage : IMimeMessage, IDisposable {
    internal static readonly Encoding DefaultCharset = Charsets.ISO8859_1;
    internal static readonly MimeType DefaultMimeType = MimeType.TextPlain;

    public MimeHeaderCollection Headers {
      get { CheckDisposed(); return headers; }
    }

    internal Stream Content {
      get { CheckDisposed(); return content; }
      set { content = value; }
    }

    public ReadOnlyCollection<MimeMessage> SubParts {
      get { CheckDisposed(); return subParts; }
    }

    public MimeMessage ParentPart {
      get { CheckDisposed(); return parentPart; }
      internal set { parentPart = value; }
    }

    public MimeMessage RootPart {
      get {
        CheckDisposed();

        var root = this;

        for (;;) {
          var p = root.ParentPart;

          if (p == null)
            return root;
          else
            root = p;
        }
      }
    }

    public MimeMessage[] AllParts {
      get { return GetFlattenSubParts().ToArray(); }
    }

    public string PartSpecifier {
      get {
        CheckDisposed();

        if (parentPart == null) {
          return string.Empty;
        }
        else {
          var ret = parentPart.PartSpecifier;

          if (0 < ret.Length)
            ret += ".";

          return ret + (parentPart.subParts.IndexOf(this) + 1).ToString();
        }
      }
    }

    public MimeType MimeType {
      get { return mimeType; }
    }

    public Encoding Charset {
      get { return charset; }
    }

    public string Boundary {
      get { return boundary; }
    }

    public ContentTransferEncodingMethod TransferEncoding {
      get { return transferEncoding; }
    }

    public MimeMessageDisposition Disposition {
      get { return disposition; }
    }

    internal MimeMessage(MimeHeaderCollection headers)
    {
      this.headers = headers;
      this.content = null;

      SetSubParts(null);
    }

    internal void SetSubParts(IEnumerable<MimeMessage> subParts)
    {
      if (subParts == null)
        this.subParts = (new List<MimeMessage>()).AsReadOnly(); // empty list
      else
        this.subParts = subParts.ToList().AsReadOnly();
    }

    internal void SetContentType(MimeType mimeType, Encoding charset, string boundary)
    {
      this.mimeType = mimeType ?? DefaultMimeType;
      this.charset = charset ?? DefaultCharset;
      this.boundary = boundary;
    }

    internal void SetContentTransferEncoding(ContentTransferEncodingMethod transferEncoding)
    {
      this.transferEncoding = transferEncoding;
    }

    internal void SetContentDisposition(MimeMessageDisposition disposition)
    {
      this.disposition = disposition;
    }

    internal IEnumerable<MimeMessage> GetFlattenSubParts()
    {
      CheckDisposed();

      var stack = new Stack<MimeMessage>();

      for (var p = this;;) {
        for (var index = p.subParts.Count - 1; 0 <= index; index--) {
          stack.Push(p.subParts[index]);
        }

        if (stack.Count == 0)
          yield break;

        p = stack.Pop();

        yield return p;
      }
    }

    public void Dispose()
    {
      if (subParts != null) {
        foreach (var subPart in subParts) {
          subPart.Dispose();
        }

        subParts = null;
      }

      if (content != null) {
        if (this == RootPart) {
          var rootContentStream = content as PartialStream;

          if (rootContentStream != null)
            // InnerStream must be FileStream
            rootContentStream.InnerStream.Close();
        }

        content.Close();
        content = null;
      }

      disposed = true;
    }

    private void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private const bool loadContentToMemoryParameterDefault = false;

    public static MimeMessage Load(string file)
    {
      return Load(file, loadContentToMemoryParameterDefault, (EncodingSelectionCallback)null, (MimeEncodedWordConverter)null);
    }

    public static MimeMessage Load(string file, bool loadContentToMemory)
    {
      return Load(file, loadContentToMemory, (EncodingSelectionCallback)null, (MimeEncodedWordConverter)null);
    }

    public static MimeMessage Load(string file,
                                   Encoding fallbackCharset)
    {
      return Load(file, loadContentToMemoryParameterDefault, fallbackCharset);
    }

    public static MimeMessage Load(string file,
                                   bool loadContentToMemory,
                                   Encoding fallbackCharset)
    {
      if (fallbackCharset == null)
        throw new ArgumentNullException("fallbackCharset");

      return Load(file, delegate(string charsetName) {
        return fallbackCharset;
      }, null);
    }

    public static MimeMessage Load(string file,
                                   EncodingSelectionCallback selectFallbackCharset)
    {
      return Load(file, loadContentToMemoryParameterDefault, selectFallbackCharset, null);
    }

    public static MimeMessage Load(string file,
                                   bool loadContentToMemory,
                                   EncodingSelectionCallback selectFallbackCharset)
    {
      return Load(file, loadContentToMemory, selectFallbackCharset, null);
    }

    public static MimeMessage Load(string file,
                                   EncodingSelectionCallback selectFallbackCharset,
                                   MimeEncodedWordConverter decodeMalformedHeaderValue)
    {
      return Load(file, loadContentToMemoryParameterDefault, selectFallbackCharset, decodeMalformedHeaderValue);
    }

    public static MimeMessage Load(string file,
                                   bool loadContentToMemory,
                                   EncodingSelectionCallback selectFallbackCharset,
                                   MimeEncodedWordConverter decodeMalformedHeaderValue)
    {
      FileStream stream = null;

      try {
        stream = File.OpenRead(file);

        return LoadCore(stream, loadContentToMemory, selectFallbackCharset, decodeMalformedHeaderValue);
      }
      catch {
        if (stream != null) {
          stream.Close();
          stream = null;
        }

        throw;
      }
      finally {
        if (loadContentToMemory && stream != null) {
          stream.Close();
          stream = null;
        }
      }
    }

    public static MimeMessage LoadMessage(string message)
    {
      return LoadMessage(message, (EncodingSelectionCallback)null, (MimeEncodedWordConverter)null);
    }

    public static MimeMessage LoadMessage(string message,
                                          Encoding fallbackCharset)
    {
      if (fallbackCharset == null)
        throw new ArgumentNullException("fallbackCharset");

      return LoadMessage(message, delegate(string charsetName) {
        return fallbackCharset;
      }, null);
    }

    public static MimeMessage LoadMessage(string message,
                                          EncodingSelectionCallback selectFallbackCharset)
    {
      return LoadMessage(message, selectFallbackCharset, null);
    }

    public static MimeMessage LoadMessage(string message,
                                          EncodingSelectionCallback selectFallbackCharset,
                                          MimeEncodedWordConverter decodeMalformedHeaderValue)
    {
      const int lengthThreshold = 20480;

      if (message == null)
        throw new ArgumentNullException("message");

      Stream messageStream = null;

      try {
        if (message.Length <= lengthThreshold) {
          messageStream = new MemoryStream(Charsets.ISO8859_1.GetBytes(message));
        }
        else {
          messageStream = new ChunkedMemoryStream();

          var writer = new StreamWriter(messageStream, Charsets.ISO8859_1);

          writer.Write(message);
          writer.Flush();
          // writer.Close(); // leave stream open

          messageStream.Position = 0L;
        }

        return LoadCore(messageStream,
                        true,
                        selectFallbackCharset,
                        decodeMalformedHeaderValue);
      }
      finally {
        if (messageStream != null)
          messageStream.Close();
      }
    }

    public static MimeMessage Load(Stream stream)
    {
      return LoadCore(stream, true, (EncodingSelectionCallback)null, (MimeEncodedWordConverter)null);
    }

    public static MimeMessage Load(Stream stream,
                                   Encoding fallbackCharset)
    {
      if (fallbackCharset == null)
        throw new ArgumentNullException("fallbackCharset");

      return LoadCore(stream, true, delegate(string charsetName) {
        return fallbackCharset;
      }, null);
    }

    public static MimeMessage Load(Stream stream,
                                   EncodingSelectionCallback selectFallbackCharset)
    {
      return LoadCore(stream, true, selectFallbackCharset, null);
    }

    public static MimeMessage Load(Stream stream,
                                   EncodingSelectionCallback selectFallbackCharset,
                                   MimeEncodedWordConverter decodeMalformedHeaderValue)
    {
      return LoadCore(stream, true, selectFallbackCharset, decodeMalformedHeaderValue);
    }

    private static MimeMessage LoadCore(Stream stream,
                                        bool loadContentToMemory,
                                        EncodingSelectionCallback selectFallbackCharset,
                                        MimeEncodedWordConverter decodeMalformedHeaderValue)
    {
      return Parser.Parse(new LooseLineOrientedStream(stream),
                          loadContentToMemory,
                          selectFallbackCharset,
                          decodeMalformedHeaderValue);
    }

    private const bool readContentDecodeParameterDefault = true;

    public void ReadContent(Action<Stream> read)
    {
      ReadContent(readContentDecodeParameterDefault, read);
    }

    public void ReadContent(bool decode, Action<Stream> read)
    {
      CheckDisposed();

      if (read == null)
        throw new ArgumentNullException("read");

      read(Decoder.CreateContentReadingStream(this, decode, true));
    }

    public void ReadContent(Action<StreamReader> read)
    {
      CheckDisposed();

      if (read == null)
        throw new ArgumentNullException("read");

      read(Decoder.CreateContentTextReader(this));
    }

    public void ReadContent(Action<BinaryReader> read)
    {
      CheckDisposed();

      if (read == null)
        throw new ArgumentNullException("read");

      read(Decoder.CreateContentBinaryReader(this));
    }

    public string ReadContentAsText()
    {
      CheckDisposed();

      return Decoder.CreateContentTextReader(this).ReadToEnd();
    }

    public byte[] ReadContentAsBinary()
    {
      CheckDisposed();

      var contentStream = Decoder.CreateContentReadingStream(this, true, false);
      var resultStream = new MemoryStream((int)contentStream.Length);

      contentStream.CopyTo(resultStream);

      resultStream.Close();

      return resultStream.ToArray();
    }

    public void CopyContentTo(Stream stream)
    {
      CopyContentTo(stream, readContentDecodeParameterDefault);
    }

    public void CopyContentTo(Stream stream, bool decode)
    {
      if (stream == null)
        throw new ArgumentNullException("stream");

      Decoder.CreateContentReadingStream(this, decode, false).CopyTo(stream);
    }

    public void SaveContent(string path)
    {
      SaveContent(new FileInfo(path), readContentDecodeParameterDefault);
    }

    public void SaveContent(string path, bool decode)
    {
      SaveContent(new FileInfo(path), decode);
    }

    internal void SaveContent(FileInfo file, bool decode)
    {
      CheckDisposed();

      using (var fileStream = file.OpenWrite()) {
        fileStream.SetLength(0L);

        Decoder.CreateContentReadingStream(this, decode, false).CopyTo(fileStream);
      }
    }
#if false
    public override string ToString()
    {
      CheckDisposed();

      // TODO: impl
      using (var builder = new MimeMessageBuilder(this)) {
        return builder.ToString();
      }
    }
#endif

    private bool disposed = false;
    private /*readonly*/ MimeMessage parentPart;
    private /*readonly*/ MimeHeaderCollection headers;
    private Stream content;
    private /*readonly*/ MimeType mimeType = DefaultMimeType;
    private /*readonly*/ Encoding charset = DefaultCharset;
    private /*readonly*/ string boundary = null;
    private /*readonly*/ ContentTransferEncodingMethod transferEncoding = ContentTransferEncodingMethod.SevenBit;
    private /*readonly*/ MimeMessageDisposition disposition = MimeMessageDisposition.None;
    private /*readonly*/ ReadOnlyCollection<MimeMessage> subParts;
  }
}