// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2012 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
#if NET_3_5
using System.Linq;
#else
using Smdn.Collections;
#endif
using System.Text;

using Smdn.IO;

namespace Smdn.Formats.Mime {
 internal static class Parser {
    private static string DefaultConvertMalformedHeaderValue(Encoding charset, string method, string text)
    {
      return null;
    }

    internal static MimeMessage Parse(LineOrientedStream stream,
                                      bool loadContentToMemory,
                                      EncodingSelectionCallback selectFallbackCharset,
                                      MimeEncodedWordConverter decodeMalformedHeaderValue)
    {
      var headers = MimeUtils.ParseHeader(stream, true).Select(delegate(KeyValuePair<string, string> pair) {
        MimeEncodingMethod encoding;
        Encoding charset;

        var headerValue = MimeEncoding.Decode(pair.Value.Trim(),
                                              selectFallbackCharset,
                                              decodeMalformedHeaderValue ?? DefaultConvertMalformedHeaderValue,
                                              out encoding,
                                              out charset);

        return new MimeHeader(pair.Key, headerValue, encoding, charset);
      });

      return ParseBody(stream,
                       loadContentToMemory,
                       new MimeHeaderCollection(headers.ToArray()),
                       selectFallbackCharset,
                       decodeMalformedHeaderValue);
    }

    private static MimeMessage ParseBody(LineOrientedStream stream,
                                         bool loadContentToMemory,
                                         MimeHeaderCollection headers,
                                         EncodingSelectionCallback selectFallbackCharset,
                                         MimeEncodedWordConverter decodeMalformedHeaderValue)
    {
      var message = new MimeMessage(headers);

      ParseContentType(message, selectFallbackCharset);
      ParseContentTransferEncoding(message);
      ParseContentDisposition(message, selectFallbackCharset);

      // read and parse content
      var baseFileStream = (Stream)(stream.InnerStream as PartialStream) ?? (Stream)(stream.InnerStream as FileStream);

      if (message.MimeType == null || !message.MimeType.TypeEquals("multipart")) {
        if (loadContentToMemory) {
          message.Content = new ChunkedMemoryStream();

          stream.CopyTo(message.Content);
        }
        else {
          message.Content = PartialStream.CreateNonNested(baseFileStream,
                                                          stream.Position,
                                                          stream.Length - stream.Position,
                                                          false);
        }

        return message;
      }

      // multipart/*
      var parts = new List<MimeMessage>();
      var delimiter = ByteString.CreateImmutable("--" + message.Boundary);
      var closeDelimiter = ByteString.CreateImmutable("--" + message.Boundary + "--");
      var offsetStart = loadContentToMemory ? 0L : stream.Position;
      Stream body = null;
      ByteString line = null;
      ByteString lastLine = null;

      var contentStream = loadContentToMemory ? new ChunkedMemoryStream() : null;

      for (;;) {
        if (loadContentToMemory && lastLine != null)
          contentStream.Write(lastLine.Segment.Array,
                              lastLine.Segment.Offset,
                              lastLine.Segment.Count);

        var offsetEnd = loadContentToMemory ? 0L : stream.Position;
        var l = stream.ReadLine();

        if (l == null)
          break;

        lastLine = line;
        line = ByteString.CreateImmutable(l);

        if (line.StartsWith(delimiter)) {
          if (lastLine != null) {
            if (ByteString.IsTerminatedByCRLF(lastLine)) {
              // CRLF "--" boundary
              if (loadContentToMemory)
                contentStream.Write(lastLine.Segment.Array,
                                    lastLine.Segment.Offset,
                                    lastLine.Segment.Count - 2);
              else
                offsetEnd -= 2;
            }
            else {
              // LF "--" boundary or CR "--" boundary
              if (loadContentToMemory)
                contentStream.Write(lastLine.Segment.Array,
                                    lastLine.Segment.Offset,
                                    lastLine.Segment.Count - 1);
              else
                offsetEnd -= 1;
            }
          }

          if (body == null) {
            if (loadContentToMemory)
              body = contentStream;
            else
              body = PartialStream.CreateNonNested(baseFileStream,
                                                   offsetStart,
                                                   offsetEnd - offsetStart,
                                                   false);
          }
          else {
            var offsetStartOfPart = loadContentToMemory ? 0L : stream.Position;
            Stream partStream;

            if (loadContentToMemory)
              partStream = contentStream;
            else
              partStream = PartialStream.CreateNonNested(baseFileStream,
                                                         offsetStart,
                                                         offsetEnd - offsetStart,
                                                         false);

            partStream.Position = 0L;

            var part = Parse(new LooseLineOrientedStream(partStream),
                             loadContentToMemory,
                             selectFallbackCharset,
                             decodeMalformedHeaderValue);

            part.ParentPart = message;

            parts.Add(part);

            if (!loadContentToMemory)
              stream.Position = offsetStartOfPart;
          }

          if (line.StartsWith(closeDelimiter))
            break;
          else if (loadContentToMemory)
            contentStream = new ChunkedMemoryStream();

          lastLine = null;
          offsetStart = loadContentToMemory ? 0L : stream.Position;
        } // StartsWith(delimiter)
      }

      message.Content = body;
      message.SetSubParts(parts);

      return message;
    }

    private static void ParseContentType(MimeMessage message,
                                         EncodingSelectionCallback selectFallbackCharset)
    {
      MimeHeader contentType;

      if (!message.Headers.TryGetHeader(MimeHeaderNames.ContentType, out contentType))
        return;

      contentType.EnsureValueParsed(selectFallbackCharset);

      MimeType mimeType = null;

      MimeType.TryParse(contentType.GetValueWithoutParameter(), out mimeType);

      Encoding charset = null;
      var charsetString = contentType.GetParameter("charset");

      if (!string.IsNullOrEmpty(charsetString))
        charset = Charsets.FromString(charsetString, selectFallbackCharset);

      message.SetContentType(mimeType, charset, contentType.GetParameter("boundary"));
    }

    private static void ParseContentTransferEncoding(MimeMessage message)
    {
      MimeHeader contentTransferEncoding;

      if (!message.Headers.TryGetHeader(MimeHeaderNames.ContentTransferEncoding, out contentTransferEncoding))
        return;

      var contentTransferEncodingString = contentTransferEncoding.GetValueWithoutParameter();

      if (string.IsNullOrEmpty(contentTransferEncodingString))
        return;

      message.SetContentTransferEncoding(ContentTransferEncoding.GetEncodingMethod(contentTransferEncodingString));
    }

    private static void ParseContentDisposition(MimeMessage message, EncodingSelectionCallback selectFallbackCharset)
    {
      MimeHeader contentDisposition;

      if (!message.Headers.TryGetHeader(MimeHeaderNames.ContentDisposition, out contentDisposition))
        return;

      contentDisposition.EnsureValueParsed(selectFallbackCharset);

      var contentDispositionString = contentDisposition.GetValueWithoutParameter();

      if (string.Equals("inline", contentDispositionString, StringComparison.OrdinalIgnoreCase))
        message.SetContentDisposition(MimeMessageDisposition.Inline);
      else if (string.Equals("attachment", contentDispositionString, StringComparison.OrdinalIgnoreCase))
        message.SetContentDisposition(MimeMessageDisposition.Attachment);
      else if (string.Equals("form-data", contentDispositionString, StringComparison.OrdinalIgnoreCase))
        message.SetContentDisposition(MimeMessageDisposition.FormData);
      else
        message.SetContentDisposition(MimeMessageDisposition.Unknown);
    }
  }
}