using System;
#if NET_3_5
using System.Linq;
#else
using Smdn.Collections;
#endif
using System.IO;
using System.Text;
using NUnit.Framework;

using Smdn.IO;

namespace Smdn.Formats.Mime {
  [TestFixture]
  public class MimeMessageTests {
    [Test/*, ExpectedException(typeof(NotSupportedException))*/]
    public void TestLoadMessageUnsupportedMimeVersion()
    {
      MimeMessage.LoadMessage(@"MIME-Version: 1.1
Content-Type: text/plain; charset=""iso-2022-jp""
Content-Transfer-Encoding: quoted-printable
");
    }

    [Test]
    public void TestLoadMessageMimeVersionNotExist()
    {
      var mime = MimeMessage.LoadMessage(@"Content-Type: text/plain");

      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.IsTrue(mime.Headers.Contains("Content-Type"));
      Assert.AreEqual("text/plain", mime.Headers["Content-Type"].Value);
    }

    [Test/*, ExpectedException(typeof(NotSupportedException))*/]
    public void TestLoadMessageMimeVersionNotSupported()
    {
      MimeMessage.LoadMessage(@"MIME-Version: 1.1\r\n");
    }

    [Test]
    // TODO: [ExpectedException(typeof(NotSupportedException))]
    public void TestLoadMessageMimeVersionInvalid()
    {
      MimeMessage.LoadMessage("MIME-Version: \"1.0 (comment)\"\r\n");
    }

    [Test]
    public void TestLoadMessageMimeVersionWithComment()
    {
      MimeMessage.LoadMessage("MIME-Version: 1.0 (comment)\r\n");
    }

    [Test]
    public void TestLoadMessageMimeVersionQuoted()
    {
      MimeMessage.LoadMessage("MIME-Version: \"1.0\"\r\n");
    }

    public enum LoadMessageMode {
      OnMemory,
      FileStream,
    }

    private void LoadMessage(string content, LoadMessageMode mode, Action<MimeMessage> action)
    {
      if (mode == LoadMessageMode.FileStream) {
        const string file = "message.eml";

        try {
          if (File.Exists(file))
            File.Delete(file);

          File.WriteAllText(file, content, Charsets.Latin1);

          using (var mime = MimeMessage.Load(file, false)) {
            if (mime.TransferEncoding != ContentTransferEncodingMethod.Unknown) {
              mime.ReadContent(delegate(Stream s) {
                switch (mime.TransferEncoding) {
                  case ContentTransferEncodingMethod.Binary:
                  case ContentTransferEncodingMethod.EightBit:
                  case ContentTransferEncodingMethod.SevenBit: {
                    Assert.IsInstanceOf<NonClosingStream>(s, "unwrapped stream #1");
                    var ncs = s as NonClosingStream;
                    Assert.IsInstanceOf<PartialStream>(ncs.InnerStream, "unwrapped stream #2");
                    var ps = ncs.InnerStream as PartialStream;
                    Assert.IsInstanceOf<FileStream>(ps.InnerStream, "unwrapped stream #3");
                    var fs = ps.InnerStream as FileStream;
                    Assert.AreEqual(file, Path.GetFileName(fs.Name), "unwrapped stream file");
                    break;
                  }

                  case ContentTransferEncodingMethod.Base64:
                  case ContentTransferEncodingMethod.QuotedPrintable: {
                    Assert.IsInstanceOf<NonClosingStream>(s, "unwrapped stream #1");
                    var ncs = s as NonClosingStream;
                    Assert.IsInstanceOf<System.Security.Cryptography.CryptoStream>(ncs.InnerStream, "unwrapped stream #2");
                    break;
                  }
                }
              });
            }

            action(mime);
          }
        }
        finally {
          if (File.Exists(file))
            File.Delete(file);
        }
      }
      else {
        using (var mime = MimeMessage.LoadMessage(content)) {
          action(mime);
        }
      }
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestLoadMessageFromStringLargeMessage(LoadMessageMode mode)
    {
      var header = "Content-Type: text/plain\r\n\r\n";
      var body = new StringBuilder();

      var line = "0123456789"; // 10

      line += line; // 20
      line += line; // 40
      line += line; // 80

      while (body.Length <= 40960) {
        body.AppendLine(line);
      }

      LoadMessage(header + body, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(body.ToString(),
                        mime.ReadContentAsText());
      });
    }

    [TestCase("\n", LoadMessageMode.OnMemory, Description = "LF, OnMemory")]
    [TestCase("\r", LoadMessageMode.OnMemory, Description = "CR, OnMemory")]
    [TestCase("\r\n", LoadMessageMode.OnMemory, Description = "CRLF, OnMemory")]
    [TestCase("\n", LoadMessageMode.FileStream, Description = "LF, FileStream")]
    [TestCase("\r", LoadMessageMode.FileStream, Description = "CR, FileStream")]
    [TestCase("\r\n", LoadMessageMode.FileStream, Description = "CRLF, FileStream")]
    public void TestLoadMessageBodyHasSinglePartTextBody(string newline, LoadMessageMode mode)
    {
      var content = @"Content-Type: text/plain
Subject: single part

single-part text body
".Replace("\r\n", "\n").Replace("\n", newline);

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.IsTrue(mime.Headers.Contains("Content-Type"));
        Assert.AreEqual("text/plain", mime.Headers["Content-Type"].Value);
        Assert.IsTrue(mime.Headers.Contains("Subject"));
        Assert.AreEqual("single part", mime.Headers["Subject"].Value);
        Assert.AreEqual("single-part text body" + newline, mime.ReadContentAsText());
      });
    }

    [TestCase("\n", LoadMessageMode.OnMemory, Description = "LF, OnMemory")]
    [TestCase("\r", LoadMessageMode.OnMemory, Description = "CR, OnMemory")]
    [TestCase("\r\n", LoadMessageMode.OnMemory, Description = "CRLF, OnMemory")]
    [TestCase("\n", LoadMessageMode.FileStream, Description = "LF, FileStream")]
    [TestCase("\r", LoadMessageMode.FileStream, Description = "CR, FileStream")]
    [TestCase("\r\n", LoadMessageMode.FileStream, Description = "CRLF, FileStream")]
    public void TestLoadMessageBodyHasMultiPartTextBody(string newline, LoadMessageMode mode)
    {
      var content = @"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain

first part, text/plain

--==boundary==
Content-Type: text/html

<html><body><p>second part, text/html</p></body></html>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", newline);

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.MultipartAlternative, mime.MimeType);
        Assert.AreEqual("==boundary==", mime.Boundary);
        Assert.IsTrue(mime.Headers.Contains("Content-Type"));
        Assert.AreEqual("multipart/alternative", mime.Headers["Content-Type"].GetValueWithoutParameter());
        Assert.AreEqual("==boundary==", mime.Headers["Content-Type"].GetParameter("boundary"));
        Assert.IsTrue(mime.Headers.Contains("Subject"));
        Assert.AreEqual("multi part", mime.Headers["Subject"].Value);
        Assert.AreEqual("multipart message body" + newline, mime.ReadContentAsText());

        Assert.AreEqual(2, mime.SubParts.Count);

        Assert.AreEqual(MimeType.TextPlain, mime.SubParts[0].MimeType);
        Assert.IsTrue(mime.SubParts[0].Headers.Contains("Content-Type"));
        Assert.AreEqual("text/plain", mime.SubParts[0].Headers["Content-Type"].Value);
        Assert.AreEqual("first part, text/plain" + newline, mime.SubParts[0].ReadContentAsText());

        Assert.AreEqual(new MimeType("text", "html"), mime.SubParts[1].MimeType);
        Assert.IsTrue(mime.SubParts[1].Headers.Contains("Content-Type"));
        Assert.AreEqual("text/html", mime.SubParts[1].Headers["Content-Type"].Value);
        Assert.AreEqual("<html><body><p>second part, text/html</p></body></html>",
                        mime.SubParts[1].ReadContentAsText());
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestLoadMessageBodyHasMultiPartWithNoHeader(LoadMessageMode mode)
    {
      var content = @"Return-Path: <>
Date: Fri, 21 Dec 2007 07:04:35 +0900
MIME-Version: 1.0
Content-Type: multipart/report; report-type=delivery-status;
  boundary=""lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx""
Subject: Warning: could not send message for past 4 hours
Auto-Submitted: auto-generated (warning-timeout)

This is a MIME-encapsulated message

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx

    **********************************************
    **      THIS IS A WARNING MESSAGE ONLY      **
    **  YOU DO NOT NEED TO RESEND YOUR MESSAGE  **
    **********************************************

The original message was received at Fri, 21 Dec 2007 02:06:24 +0900
from xxx.xxxxxx.xxx [127.0.0.1]

   ----- Transcript of session follows -----
451 xxx.xxxxxx.xxx: Name server timeout
Warning: message still undelivered after 4 hours
Will keep trying until message is 5 days old
451 xxx.xxxxxx.xxx: Name server timeout

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx
Content-Type: message/delivery-status

Reporting-MTA: dns; xxx.xxxxxx.xxx
Arrival-Date: Fri, 21 Dec 2007 02:06:24 +0900

Final-Recipient: RFC822; root@xxx.xxxxxx.xxx
X-Actual-Recipient: RFC822; xxxxxx@xxx.xxxxxx.xxx
Action: delayed
Status: 4.4.3
Last-Attempt-Date: Fri, 21 Dec 2007 07:04:35 +0900
Will-Retry-Until: Wed, 26 Dec 2007 02:06:24 +0900

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx
Content-Type: message/rfc822

Return-Path: <root@xxx.xxxxxx.xxx>
Date: Fri, 21 Dec 2007 02:02:15 +0900
Content-Type: text/plain; charset=UTF-8
Auto-Submitted: auto-generated
X-Cron-Env: <MAILTO=root,postmaster,webmaster,clamav>
X-Cron-Env: <SHELL=/bin/sh>
X-Cron-Env: <HOME=/root>
X-Cron-Env: <PATH=/usr/bin:/bin>
X-Cron-Env: <LOGNAME=root>
X-Cron-Env: <USER=root>

ERROR: Can't query current.cvd.clamav.net
ERROR: Can't get information about database.clamav.net: Temporary DNS error
ERROR: No servers could be reached. Giving up
ERROR: Can't query current.cvd.clamav.net
ERROR: Can't get information about database.clamav.net: Temporary DNS error
ERROR: No servers could be reached. Giving up
ERROR: Can't query current.cvd.clamav.net
ERROR: Can't get information about database.clamav.net: Temporary DNS error
ERROR: No servers could be reached. Giving up
ERROR: Update failed. Your network may be down or none of the mirrors listed in freshclam.conf is working.

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx--
";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(new MimeType("multipart", "report"), mime.MimeType);
        Assert.AreEqual("lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx", mime.Boundary);
        Assert.IsTrue(mime.Headers.Contains("Content-Type"));
        Assert.AreEqual("multipart/report", mime.Headers["Content-Type"].GetValueWithoutParameter());
        Assert.AreEqual("lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx", mime.Headers["Content-Type"].GetParameter("boundary"));

        Assert.AreEqual(3, mime.SubParts.Count);

        Assert.AreEqual(0, mime.SubParts[0].Headers.Count);
        Assert.IsTrue(mime.SubParts[0].ReadContentAsText().StartsWith("    ***"));

        Assert.AreEqual(new MimeType("message", "delivery-status"), mime.SubParts[1].MimeType);
        Assert.AreEqual("message/delivery-status", mime.SubParts[1].Headers["Content-Type"].Value);
        Assert.AreEqual(new MimeType("message", "rfc822"), mime.SubParts[2].MimeType);
        Assert.AreEqual("message/rfc822", mime.SubParts[2].Headers["Content-Type"].Value);
      });
    }

    private class NonSeekableMemoryStream : MemoryStream {
      public NonSeekableMemoryStream(byte[] buffer)
        : base(buffer)
      {
      }

      public override bool CanSeek {
        get { return false; }
      }

      public override long Position {
        get { throw new NotSupportedException(); }
        set { throw new NotSupportedException(); }
      }

      public override long Length {
        get { throw new NotSupportedException(); }
      }

      public override void SetLength(long value)
      {
        throw new NotSupportedException();
      }

      public override long Seek(long offset, SeekOrigin loc)
      {
        throw new NotSupportedException();
      }
    }

    [Test]
    public void TestLoadMessageFromNonSeekableStream()
    {
      var stream = new NonSeekableMemoryStream(Charsets.Latin1.GetBytes(@"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain

first part, text/plain

--==boundary==
Content-Type: text/html

<html><body><p>second part, text/html</p></body></html>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n")));
      var mime = MimeMessage.Load(stream);

      Assert.AreEqual("multipart message body\r\n", mime.ReadContentAsText());

      Assert.AreEqual(2, mime.SubParts.Count);
      
      Assert.AreEqual("first part, text/plain\r\n", mime.SubParts[0].ReadContentAsText());
      
      Assert.AreEqual("<html><body><p>second part, text/html</p></body></html>",
                      mime.SubParts[1].ReadContentAsText());
    }

    private void UsingTestFile(string content, Action<FileInfo> testAction)
    {
      const string file = "test.eml";

      try {
        if (File.Exists(file))
          File.Delete(file);

        File.WriteAllText(file, content, Charsets.Latin1);

        testAction(new FileInfo(file));
      }
      finally {
        if (File.Exists(file))
          File.Delete(file);
      }
    }

    [Test]
    public void TestLoadMessageNonBufferedSinglePart()
    {
      var message = @"Content-Type: text/plain
Subject: single part

<content>".Replace("\r\n", "\n").Replace("\n", "\r\n");

      UsingTestFile(message, delegate(FileInfo file) {
        using (var mime = MimeMessage.Load(file.FullName, false)) {
          mime.ReadContent(delegate(Stream stream) {
            Assert.IsInstanceOf<NonClosingStream>(stream);

            var ncs = stream as NonClosingStream;

            Assert.IsInstanceOf<PartialStream>(ncs.InnerStream);

            var ps = ncs.InnerStream as PartialStream;

            Assert.IsInstanceOf<FileStream>(ps.InnerStream);

            var fs = ps.InnerStream as FileStream;

            var offset = message.IndexOf("<content>");

            Assert.AreEqual(offset, fs.Position);
            Assert.AreEqual("<content>".Length, stream.Length);
          });


          Assert.AreEqual("<content>",
                          mime.ReadContentAsText());
        }
      });
    }

    [Test]
    public void TestLoadMessageNonBufferedMultiPart()
    {
      var message = @"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/plain

<part#0>
--==boundary==
Content-Type: text/plain

<part#1>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      UsingTestFile(message, delegate(FileInfo file) {
        using (var mime = MimeMessage.Load(file.FullName, false)) {
          mime.ReadContent(delegate(Stream stream) {
            Assert.IsInstanceOf<NonClosingStream>(stream, "<content>");

            var ncs = stream as NonClosingStream;

            Assert.IsInstanceOf<PartialStream>(ncs.InnerStream, "<content>");

            var ps = ncs.InnerStream as PartialStream;

            Assert.IsInstanceOf<FileStream>(ps.InnerStream, "<content>");

            var fs = ps.InnerStream as FileStream;

            var offset = message.IndexOf("<content>");

            Assert.AreEqual(offset, fs.Position, "position of <content>");
            Assert.AreEqual("<content>".Length, stream.Length, "length of <content>");
          });

          Assert.AreEqual("<content>",
                          mime.ReadContentAsText(), "content of <content>");

          mime.SubParts[0].ReadContent(delegate(Stream stream) {
            Assert.IsInstanceOf<NonClosingStream>(stream, "<part#0>");

            var ncs = stream as NonClosingStream;

            Assert.IsInstanceOf<PartialStream>(ncs.InnerStream, "<part#0>");

            var ps = ncs.InnerStream as PartialStream;

            Assert.IsInstanceOf<FileStream>(ps.InnerStream, "<part#0>");

            var fs = ps.InnerStream as FileStream;

            var offset = message.IndexOf("<part#0>");

            Assert.AreEqual(offset, fs.Position, "position of <part#0>");
            Assert.AreEqual("<part#0>".Length, stream.Length, "length of<part#0>");
          });

          Assert.AreEqual("<part#0>",
                          mime.SubParts[0].ReadContentAsText(), "content of <part#0>");

          mime.SubParts[1].ReadContent(delegate(Stream stream) {
            Assert.IsInstanceOf<NonClosingStream>(stream, "<part#1>");

            var ncs = stream as NonClosingStream;

            Assert.IsInstanceOf<PartialStream>(ncs.InnerStream, "<part#1>");

            var ps = ncs.InnerStream as PartialStream;

            Assert.IsInstanceOf<FileStream>(ps.InnerStream, "<part#1>");

            var fs = ps.InnerStream as FileStream;

            var offset = message.IndexOf("<part#1>");

            Assert.AreEqual(offset, fs.Position, "position of <part#1>");
            Assert.AreEqual("<part#1>".Length, stream.Length, "length of <part#1>");
          });

          Assert.AreEqual("<part#1>",
                          mime.SubParts[1].ReadContentAsText(), "content of <part#1>");
        }
      });
    }

    [Test]
    public void TestLoadMessageNonBufferedMultiPartDecode()
    {
      var message = @"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Content-Transfer-Encoding: base64
Subject: multi part

PGNvbnRlbnQ+
--==boundary==
Content-Type: text/plain
Content-Transfer-Encoding: base64

PHBhcnQjMD4=
--==boundary==
Content-Type: text/plain
Content-Transfer-Encoding: quoted-printable

=3Cpart=231=3E
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      UsingTestFile(message, delegate(FileInfo file) {
        using (var mime = MimeMessage.Load(file.FullName, false)) {
          Assert.AreEqual("<content>",
                          mime.ReadContentAsText(), "content of <content>");
          Assert.AreEqual("<part#0>",
                          mime.SubParts[0].ReadContentAsText(), "content of <part#0>");
          Assert.AreEqual("<part#1>",
                          mime.SubParts[1].ReadContentAsText(), "content of <part#1>");
        }
      });
    }

    [Test]
    public void TestLoadMessageNonBufferedMultiPartNested()
    {
      var message = @"Content-Type: multipart/mixed;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/plain

<part#0>
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

<nested>
--==nested==
Content-Type:text/plain

<nested#0>
--==nested==
Content-Type:text/plain

<nested#1>
--==nested==--
--==boundary==
Content-Type: text/plain

<part#1>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      UsingTestFile(message, delegate(FileInfo file) {
        using (var mime = MimeMessage.Load(file.FullName, false)) {
          mime.ReadContent(delegate(Stream stream) {
            Assert.DoesNotThrow(delegate {
              var ncs = stream as NonClosingStream;
              var ps = ncs.InnerStream as PartialStream;
              var fs = ps.InnerStream as FileStream;
              
              Assert.IsNotNull(fs);
            });
          });

          Assert.AreEqual("<content>",
                          mime.ReadContentAsText(), "content of <content>");
          Assert.AreEqual("<part#0>",
                          mime.SubParts[0].ReadContentAsText(), "content of <part#0>");
          Assert.AreEqual("<nested>",
                          mime.SubParts[1].ReadContentAsText(), "content of <nested>");
          Assert.AreEqual("<nested#0>",
                          mime.SubParts[1].SubParts[0].ReadContentAsText(), "content of <nested#0>");
          Assert.AreEqual("<nested#1>",
                          mime.SubParts[1].SubParts[1].ReadContentAsText(), "content of <nested#1>");
          Assert.AreEqual("<part#1>",
                          mime.SubParts[2].ReadContentAsText(), "content of <part#1>");
        }
      });
    }

    [Test]
    public void TestLoadMessageNonBufferedExceptionThrownFromParser()
    {
      var message = @"Content-Type: text/plain; charset=x-foo
Subject: single part

<content>".Replace("\r\n", "\n").Replace("\n", "\r\n");

      UsingTestFile(message, delegate(FileInfo file) {
        var ex = Assert.Throws<NotSupportedException>(delegate {
          MimeMessage.Load(file.FullName, false, delegate(string name) {
            throw new NotSupportedException("unsupported operation");
          });
        });

        Assert.AreEqual("unsupported operation", ex.Message);
      });
    }

    [Test]
    public void TestHeaders()
    {
      var message =
        "content-type: text/plain\r\n" +
          "FROM: from@example.com\r\n" + 
          "tO: to@example.com\r\n";

      var mime = MimeMessage.LoadMessage(message);

      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.IsTrue(mime.Headers.Contains("CONTENT-TYPE"));
      Assert.AreEqual("text/plain", mime.Headers["Content-Type"].Value);
      Assert.AreSame(mime.Headers["CONTENT-TYPE"], mime.Headers["Content-Type"]);

      Assert.IsTrue(mime.Headers.Contains("from"));
      Assert.AreEqual("from@example.com", mime.Headers["FROM"].Value);

      Assert.IsTrue(mime.Headers.Contains("to"));
      Assert.AreEqual("to@example.com", mime.Headers["TO"].Value);
    }

    [Test]
    public void TestHeadersMultipleHeader()
    {
      var message = @"MIME-Version: 1.0\r\n
Received: #1
Received: #2
Received: #3";

      var mime = MimeMessage.LoadMessage(message);

      Assert.IsTrue(mime.Headers.Contains("Received"));

      Assert.AreEqual(3, mime.Headers.GetHeaders("Received").Count());
      Assert.AreEqual(1, mime.Headers.GetHeaders("Received").Where(h => h.Value == "#1").Count(), "#1");
      Assert.AreEqual(1, mime.Headers.GetHeaders("Received").Where(h => h.Value == "#2").Count(), "#2");
      Assert.AreEqual(1, mime.Headers.GetHeaders("Received").Where(h => h.Value == "#3").Count(), "#2");
    }

    [Test]
    public void TestParentPart()
    {
      var message = @"Content-Type: multipart/mixed;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/plain

<part#0>
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

<nested>
--==nested==
Content-Type:text/plain

<nested#0>
--==nested==
Content-Type:text/plain

<nested#1>
--==nested==--
--==boundary==
Content-Type: text/plain

<part#1>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mime = MimeMessage.LoadMessage(message);

      Assert.IsNull(mime.ParentPart, "root");
      Assert.AreSame(mime, mime.SubParts[0].ParentPart, "root.0");
      Assert.AreSame(mime, mime.SubParts[1].ParentPart, "root.1");
      Assert.AreSame(mime.SubParts[1], mime.SubParts[1].SubParts[0].ParentPart, "root.1.0");
      Assert.AreSame(mime.SubParts[1], mime.SubParts[1].SubParts[1].ParentPart, "root.1.1");
      Assert.AreSame(mime, mime.SubParts[2].ParentPart, "root.2");
    }

    [Test]
    public void TestRootPart()
    {
      var message = @"Content-Type: multipart/mixed;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/plain

<part#0>
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

<nested>
--==nested==
Content-Type:text/plain

<nested#0>
--==nested==
Content-Type:text/plain

<nested#1>
--==nested==--
--==boundary==
Content-Type: text/plain

<part#1>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mime = MimeMessage.LoadMessage(message);

      Assert.AreSame(mime, mime.RootPart, "root");
      Assert.AreSame(mime, mime.SubParts[0].RootPart, "root.0");
      Assert.AreSame(mime, mime.SubParts[1].RootPart, "root.1");
      Assert.AreSame(mime, mime.SubParts[1].SubParts[0].RootPart, "root.1.0");
      Assert.AreSame(mime, mime.SubParts[1].SubParts[1].RootPart, "root.1.1");
      Assert.AreSame(mime, mime.SubParts[2].RootPart, "root.2");
    }

    [Test]
    public void TestAllParts()
    {
      var message = @"Content-Type: multipart/mixed;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/plain

<part#0>
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

<nested>
--==nested==
Content-Type:text/plain

<nested#0>
--==nested==
Content-Type:text/plain

<nested#1>
--==nested==--
--==boundary==
Content-Type: text/plain

<part#1>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mime = MimeMessage.LoadMessage(message);
      var parts = mime.AllParts;

      Assert.AreEqual(5, parts.Length);
      Assert.AreSame(mime.SubParts[0], parts[0], "root.0");
      Assert.AreSame(mime.SubParts[1], parts[1], "root.1");
      Assert.AreSame(mime.SubParts[1].SubParts[0], parts[2], "root.1.0");
      Assert.AreSame(mime.SubParts[1].SubParts[1], parts[3], "root.1.1");
      Assert.AreSame(mime.SubParts[2], parts[4], "root.2");
    }

    [Test]
    public void TestPartSpecifier()
    {
      var message = @"Content-Type: multipart/mixed;
  boundary=""==boundary==""
Subject: multi part

<content>
--==boundary==
Content-Type: text/plain

<part#0>
--==boundary==
Content-Type: multipart/mixed; boundary=""==nested==""

<nested>
--==nested==
Content-Type:text/plain

<nested#0>
--==nested==
Content-Type:text/plain

<nested#1>
--==nested==--
--==boundary==
Content-Type: text/plain

<part#1>
--==boundary==--".Replace("\r\n", "\n").Replace("\n", "\r\n");

      var mime = MimeMessage.LoadMessage(message);

      Assert.AreEqual(string.Empty, mime.PartSpecifier);
      Assert.AreEqual("1", mime.SubParts[0].PartSpecifier, "root.0");
      Assert.AreEqual("2", mime.SubParts[1].PartSpecifier, "root.1");
      Assert.AreEqual("2.1", mime.SubParts[1].SubParts[0].PartSpecifier, "root.1.0");
      Assert.AreEqual("2.2", mime.SubParts[1].SubParts[1].PartSpecifier, "root.1.1");
      Assert.AreEqual("3", mime.SubParts[2].PartSpecifier, "root.2");
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestDispose(LoadMessageMode mode)
    {
      var content = @"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain

first part, text/plain

--==boundary==
Content-Type: text/html

<html><body><p>second part, text/html</p></body></html>
--==boundary==--";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        var part1 = mime.SubParts[0];
        var part2 = mime.SubParts[1];

        mime.Dispose();

        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.Headers));
        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.SubParts));
        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.ParentPart));
        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.RootPart));
        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.AllParts));
        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.PartSpecifier));
        Assert.Throws<ObjectDisposedException>(() => mime.ReadContent(delegate(StreamReader r) {
          r.ReadLine();
        }));
        Assert.Throws<ObjectDisposedException>(() => mime.ReadContent(delegate(BinaryReader r) {
          r.ReadByte();
        }));
        Assert.Throws<ObjectDisposedException>(() => mime.ReadContentAsText());
        Assert.Throws<ObjectDisposedException>(() => mime.ReadContentAsBinary());
        Assert.Throws<ObjectDisposedException>(() => mime.CopyContentTo(Stream.Null));
        Assert.Throws<ObjectDisposedException>(() => mime.SaveContent("content.tmp"));

        Assert.Throws<ObjectDisposedException>(() => part1.ReadContentAsText());
        Assert.Throws<ObjectDisposedException>(() => part2.ReadContentAsText());

        Assert.DoesNotThrow(() => mime.Dispose(), "re-Dispose");
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestDisposeSubPart(LoadMessageMode mode)
    {
      var content = @"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain

first part, text/plain

--==boundary==
Content-Type: text/html

<html><body><p>second part, text/html</p></body></html>
--==boundary==--";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        var part1 = mime.SubParts[0];
        var part2 = mime.SubParts[1];

        part1.Dispose();

        Assert.Throws<ObjectDisposedException>(() => part1.ReadContentAsText());

        Assert.DoesNotThrow(() => mime.ReadContentAsText());
        Assert.DoesNotThrow(() => part2.ReadContentAsText());

        part2.Dispose();

        Assert.Throws<ObjectDisposedException>(() => part2.ReadContentAsText());

        Assert.DoesNotThrow(() => mime.ReadContentAsText());

        Assert.DoesNotThrow(() => mime.Dispose(), "Dispose root");
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentAsText(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual("content", mime.ReadContentAsText());
        Assert.AreEqual("content", mime.ReadContentAsText(), "read again");
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentAsBinary(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(Charsets.Latin1.GetBytes("content"), mime.ReadContentAsBinary());
        Assert.AreEqual(Charsets.Latin1.GetBytes("content"), mime.ReadContentAsBinary(), "read again");
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTypeNotExist(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Subject: transfer encoding not exist

decode as text/plain; charset=us-ascii";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual("decode as text/plain; charset=us-ascii", mime.ReadContentAsText());
      });
    }
    
    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTypeNonText(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: application/octet-stream
Subject: transfer encoding not exist

decode as binary stream";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        var text = Encoding.ASCII.GetBytes("decode as binary stream");

        Assert.AreEqual(MimeType.ApplicationOctetStream, mime.MimeType);
        Assert.AreEqual(text, mime.ReadContentAsBinary());
      });
    }
    
    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncoding7BitMessage(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit
Subject: 7bit transfer encoding

7bit transfer encoding";
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.AreEqual(Charsets.ASCII, mime.Charset);
        Assert.AreEqual(ContentTransferEncodingMethod.SevenBit, mime.TransferEncoding);
        Assert.AreEqual("7bit", mime.Headers["Content-Transfer-Encoding"].Value);
        Assert.AreEqual("7bit transfer encoding", mime.ReadContentAsText());
      });
    }
    
    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncoding8BitMessage(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain; charset=iso-8859-1
Content-Transfer-Encoding: 8bit
Subject: 8bit transfer encoding

" + "8bit transfer encoding \u0080 \u00cf \u00ff\r\n";
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.AreEqual(Charsets.Latin1, mime.Charset);
        Assert.AreEqual(ContentTransferEncodingMethod.EightBit, mime.TransferEncoding);
        Assert.AreEqual("8bit", mime.Headers["Content-Transfer-Encoding"].Value);
        Assert.AreEqual("8bit transfer encoding \u0080 \u00cf \u00ff\r\n", mime.ReadContentAsText());
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncodingBinaryMessage(LoadMessageMode mode)
    {
      var message = new MemoryStream(0x200);

      var header = Charsets.Latin1.GetBytes(@"MIME-Version: 1.0
Content-Type: application/octet-stream
Content-Transfer-Encoding: binary
Subject: binary transfer encoding

");

      message.Write(header, 0, header.Length);

      var bodyStream = new MemoryStream(); 

      for (var i = 0; i < 0x100; i++) {
        bodyStream.WriteByte((byte)i);
      }

      var body = bodyStream.ToArray();

      message.Write(body, 0, body.Length);

      var content = Charsets.Latin1.GetString(message.ToArray());

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.ApplicationOctetStream, mime.MimeType);
        Assert.AreEqual(ContentTransferEncodingMethod.Binary, mime.TransferEncoding);
        Assert.AreEqual("binary", mime.Headers["Content-Transfer-Encoding"].Value);

        Assert.AreEqual(body, mime.ReadContentAsBinary());
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncodingBase64Message(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain; charset=shift_jis
Content-Transfer-Encoding: base64
Subject: base64 transfer encoding

ir+OmmFiY4KpgsgxMjODSoNpDQqCUIJRglKCgYKCgoMNCg==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.AreEqual(Charsets.ShiftJIS, mime.Charset);
        Assert.AreEqual(ContentTransferEncodingMethod.Base64, mime.TransferEncoding);
        Assert.AreEqual("base64", mime.Headers["Content-Transfer-Encoding"].Value);
        Assert.AreEqual("漢字abcかな123カナ\r\n１２３ａｂｃ\r\n", mime.ReadContentAsText());
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncodingQuotedPrintableMessage(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain; charset=euc-jp
Content-Transfer-Encoding: quoted-printable
Subject: quoted printable transfer encoding

=B4=C1=BB=FAabc=A4=AB=A4=CA123=A5=AB=A5=CA=0D=0A=A3=B1=A3=B2=A3=B3=A3=E1=A3=
=E2=A3=E3=0D=0A";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.AreEqual(Encoding.GetEncoding("euc-jp"), mime.Charset);
        Assert.AreEqual(ContentTransferEncodingMethod.QuotedPrintable, mime.TransferEncoding);
        Assert.AreEqual("quoted-printable", mime.Headers["Content-Transfer-Encoding"].Value);
        Assert.AreEqual("漢字abcかな123カナ\r\n１２３ａｂｃ\r\n", mime.ReadContentAsText());
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentContentTransferEncodingUnsupported(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain; charset=euc-jp
Content-Transfer-Encoding: unsupported
Subject: unsupported transfer encoding

unsupported transfer encoding";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
        Assert.AreEqual(Encoding.GetEncoding("euc-jp"), mime.Charset);
        Assert.AreEqual(ContentTransferEncodingMethod.Unknown, mime.TransferEncoding);
        Assert.AreEqual("unsupported", mime.Headers["Content-Transfer-Encoding"].Value);

        try {
          mime.ReadContentAsText();

          Assert.Fail("NotSupportedException not thrown");
        }
        catch (NotSupportedException) {
        }
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStream(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(false, delegate(Stream contentStream) {
          using (var stream = new MemoryStream()) {
            contentStream.CopyTo(stream);

            Assert.AreEqual("Y29udGVudA==", Encoding.ASCII.GetString(stream.ToArray()));
          }
        });

        mime.ReadContent(false, delegate(Stream contentStream) {
          using (var stream = new MemoryStream()) {
            contentStream.CopyTo(stream);

            Assert.AreEqual("Y29udGVudA==", Encoding.ASCII.GetString(stream.ToArray()), "read again");
          }
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStreamDecode(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(true, delegate(Stream contentStream) {
          using (var stream = new MemoryStream()) {
            contentStream.CopyTo(stream);

            Assert.AreEqual("content", Encoding.ASCII.GetString(stream.ToArray()));
          }
        });

        mime.ReadContent(true, delegate(Stream contentStream) {
          using (var stream = new MemoryStream()) {
            contentStream.CopyTo(stream);

            Assert.AreEqual("content", Encoding.ASCII.GetString(stream.ToArray()), "read again");
          }
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStreamCloseStreamPlainEncoding(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(false, delegate(Stream contentStream) {
          contentStream.Close();
        });

        mime.ReadContent(false, delegate(Stream contentStream) {
          using (var stream = new MemoryStream()) {
            contentStream.CopyTo(stream);

            Assert.AreEqual("content", Encoding.ASCII.GetString(stream.ToArray()), "read again");
          }
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStreamCloseStreamNonPlainEncoding(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(true, delegate(Stream contentStream) {
          contentStream.Close();
        });

        mime.ReadContent(true, delegate(Stream contentStream) {
          using (var stream = new MemoryStream()) {
            contentStream.CopyTo(stream);

            Assert.AreEqual("content", Encoding.ASCII.GetString(stream.ToArray()), "read again");
          }
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStreamReader(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(delegate(StreamReader reader) {
          Assert.AreEqual(Charsets.Latin1, reader.CurrentEncoding);
          Assert.AreEqual("content", reader.ReadToEnd());
        });

        mime.ReadContent(delegate(StreamReader reader) {
          Assert.AreEqual("content", reader.ReadToEnd(), "read again");
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentBinaryReader(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(delegate(BinaryReader reader) {
          Assert.AreEqual(Charsets.Latin1.GetBytes("content"),
                          reader.ReadBytes(7));
          Assert.AreEqual(-1, reader.PeekChar());
        });

        mime.ReadContent(delegate(BinaryReader reader) {
          Assert.AreEqual(Charsets.Latin1.GetBytes("content"),
                          reader.ReadBytes(7),
                          "read again");
          Assert.AreEqual(-1, reader.PeekChar());
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStreamReaderCloseReaderPlainEncoding(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(delegate(StreamReader reader) {
          reader.Close();
        });

        mime.ReadContent(delegate(StreamReader reader) {
          Assert.AreEqual("content", reader.ReadToEnd(), "read again");
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentStreamReaderCloseReaderNonPlainEncoding(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(delegate(StreamReader reader) {
          reader.Close();
        });

        mime.ReadContent(delegate(StreamReader reader) {
          Assert.AreEqual("content", reader.ReadToEnd(), "read again");
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentBinaryReaderCloseReaderPlainEncoding(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(delegate(BinaryReader reader) {
          reader.Close();
        });

        mime.ReadContent(delegate(BinaryReader reader) {
          Assert.AreEqual(Charsets.Latin1.GetBytes("content"),
                          reader.ReadBytes(7),
                          "read again");
          Assert.AreEqual(-1, reader.PeekChar());
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestReadContentBinaryReaderCloseReaderNonPlainEncoding(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        mime.ReadContent(delegate(BinaryReader reader) {
          reader.Close();
        });

        mime.ReadContent(delegate(BinaryReader reader) {
          Assert.AreEqual(Charsets.Latin1.GetBytes("content"),
                          reader.ReadBytes(7),
                          "read again");
          Assert.AreEqual(-1, reader.PeekChar());
        });
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestCopyContentTo(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        using (var outStream = new MemoryStream()) {
          mime.CopyContentTo(outStream, false);

          Assert.AreEqual(Encoding.ASCII.GetBytes("Y29udGVudA=="),
                          outStream.ToArray());
        }
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestCopyContentToDecodeContent(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        using (var outStream = new MemoryStream()) {
          mime.CopyContentTo(outStream, true);

          Assert.AreEqual(Encoding.ASCII.GetBytes("content"),
                          outStream.ToArray());
        }
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestSaveContent(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        const string file = "content.txt";

        try {
          if (File.Exists(file))
            File.Delete(file);

          mime.SaveContent(file, false);

          Assert.IsTrue(File.Exists(file));
          Assert.AreEqual("Y29udGVudA==",
                          File.ReadAllText(file, Encoding.ASCII));
        }
        finally {
          if (File.Exists(file))
            File.Delete(file);
        }
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestSaveContentOverwriteExistingFile(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";
      
      LoadMessage(content, mode, delegate(MimeMessage mime) {
        const string file = "content.txt";

        try {
          if (File.Exists(file))
            File.Delete(file);

          File.WriteAllText(file, new string('*', 1024));

          mime.SaveContent(file, false);

          Assert.IsTrue(File.Exists(file));
          Assert.AreEqual("Y29udGVudA==",
                          File.ReadAllText(file, Encoding.ASCII));
        }
        finally {
          if (File.Exists(file))
            File.Delete(file);
        }
      });
    }

    [TestCase(LoadMessageMode.OnMemory)]
    [TestCase(LoadMessageMode.FileStream)]
    public void TestSaveContentDecodeContent(LoadMessageMode mode)
    {
      var content = @"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==";

      LoadMessage(content, mode, delegate(MimeMessage mime) {
        const string file = "content.txt";

        try {
          if (File.Exists(file))
            File.Delete(file);

          mime.SaveContent(file, true);

          Assert.IsTrue(File.Exists(file));
          Assert.AreEqual("content",
                          File.ReadAllText(file, Encoding.ASCII));
        }
        finally {
          if (File.Exists(file))
            File.Delete(file);
        }
      });
    }
  }
}